# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from hashlib import sha1
from itertools import chain
from logging import getLogger
from pathlib import Path
from random import shuffle
from re import compile as re_compile
from shutil import rmtree
from string import punctuation
from tempfile import gettempdir
from time import gmtime, perf_counter, sleep, strftime
from typing import Any, NewType
from urllib.parse import quote, urlsplit

from ffpuppet import BrowserTimeoutError, Debugger, FFPuppet, LaunchError, Reason

from .explorer import Explorer
from .reporter import FuzzManagerReporter

LOG = getLogger(__name__)
TMP_PATH = Path(gettempdir()) / "site-scout"
TMP_PATH.mkdir(exist_ok=True, parents=True)


# This is used as a placeholder for empty subdomains
# WARNING: If this changes all yml files will need to be updated
NO_SUBDOMAIN = "*"

# Note: Python 3.10+ use TypeAlias
UrlDB = NewType("UrlDB", dict[str, dict[str, list[str]]])


def trim(in_str: str, max_len: int) -> str:
    """Trim in_str if it exceeds max_len.

    Args:
        in_str: String to trim if required.
        max_len: Maximum length of output.

    Returns:
        Modified string or original string depending on max_len.
    """
    return f"{in_str[:max_len - 3]}..." if len(in_str) > max_len else in_str


# pylint: disable=too-many-instance-attributes
@dataclass(eq=False)
class VisitSummary:
    """Store data from completed Visits for analysis."""

    duration: float
    url: URL
    get_duration: float | None = None
    explore_duration: float | None = None
    explore_state: str | None = None
    force_closed: bool = False
    has_result: bool = False
    not_found: bool = False


class Status:
    """Track and report status to a file."""

    __slots__ = ("_dst", "_next", "_rate_limit", "_start")

    def __init__(self, dst: Path, rate_limit: int = 60) -> None:
        assert dst
        assert rate_limit >= 0
        self._dst = dst
        self._next: float = 0
        self._rate_limit = rate_limit
        self._start = perf_counter()

    def report(
        self,
        active: int,
        jobs: int,
        completed: int,
        target: int,
        results: int,
        not_found: int,
        avg_duration: int,
        force: bool = False,
    ) -> None:
        """Write status report to the filesystem.

        Args:
            active: Number of active browsers.
            jobs: Maximum number of active browsers.
            completed: URLs opened.
            target: Total URLs to be opened.
            results: Number of results found.
            not_found: Number of 'server not found' errors.
            force: Ignore rate limit and report.

        Returns:
            None
        """
        now = perf_counter()
        if not force and self._next > now:
            return
        comp_pct = (completed / target) * 100 if target else 0.0
        with self._dst.open("w") as lfp:
            lfp.write(f" Active/Limit : {active}/{jobs}\n")
            lfp.write(f"Current/Total : {completed}/{target} ({comp_pct:0.1f}%)\n")
            lfp.write(f"      Results : {results}\n")
            if not_found > 0:
                lfp.write(f"    Not Found : {not_found}\n")
            if avg_duration > 0:
                lfp.write(f" Avg Duration : {avg_duration}s\n")
            lfp.write(f"      Runtime : {timedelta(seconds=int(now - self._start))}\n")
            lfp.write(f"    Timestamp : {strftime('%Y/%m/%d %X %z', gmtime())}\n")
        self._next = now + self._rate_limit


class URL:
    """URL components."""

    ALLOWED_SCHEMES = frozenset(("http", "https"))
    VALID_DOMAIN = re_compile(r"[a-zA-Z0-9_.-]")

    __slots__ = ("_uid", "domain", "path", "scheme", "subdomain")

    def __init__(
        self,
        domain: str,
        subdomain: str | None = None,
        path: str = "/",
        scheme: str = "http",
    ) -> None:
        self.domain = domain
        self.path = path
        self.scheme = scheme
        self.subdomain = subdomain
        self._uid: str | None = None

    def __str__(self) -> str:
        if self.subdomain is None or self.subdomain == NO_SUBDOMAIN:
            return f"{self.scheme}://{self.domain}{self.path}"
        return f"{self.scheme}://{self.subdomain}.{self.domain}{self.path}"

    @staticmethod
    def is_ascii(data: str) -> bool:
        """Check if all characters in a string are ASCII.

        Args:
            data: String to check.

        Returns:
            True if string only contains ASCII characters otherwise False.
        """
        return all(ord(x) < 128 for x in data)

    # pylint: disable=too-many-return-statements
    @classmethod
    def create(
        cls,
        domain: str,
        subdomain: str | None = None,
        path: str = "/",
        scheme: str = "http",
    ) -> URL | None:
        """Sanitize, verify data and create a URL if possible.

        Args:
            domain: Domain.
            subdomain: Subdomain.
            path: Path, must begin with '/'.
            scheme: Scheme.

        Returns:
            URL object if input is valid otherwise None.
        """
        scheme = scheme.lower()
        if scheme not in cls.ALLOWED_SCHEMES:
            LOG.error("Cannot create URL: Invalid scheme %r", scheme)
            return None

        try:
            # use idna to encode domain with non ascii characters
            domain = domain.lower().encode("idna").decode("ascii")
        except UnicodeError:
            LOG.error("Cannot create URL: Invalid domain %r", domain)
            return None
        if not cls.VALID_DOMAIN.match(domain):
            LOG.error("Cannot create URL: Invalid domain %r", domain)
            return None

        if subdomain is not None and subdomain != NO_SUBDOMAIN:
            if not subdomain:
                LOG.error("Cannot create URL: Empty subdomain")
                return None
            try:
                subdomain = subdomain.lower().encode("idna").decode("ascii")
            except UnicodeError:
                LOG.error("Cannot create URL: Invalid subdomain %r", subdomain)
                return None
            if not cls.VALID_DOMAIN.match(subdomain):
                LOG.error("Cannot create URL: Invalid subdomain %r", subdomain)
                return None

        if not path.startswith("/"):
            LOG.error("Cannot create URL: Path must begin with '/'")
            return None

        # percent encode non ascii characters in path if needed
        if not cls.is_ascii(path):
            path = quote(path, safe=punctuation)

        return cls(domain, subdomain=subdomain, path=path, scheme=scheme)

    @property
    def uid(self) -> str:
        """Unique ID.

        Args:
            None

        Returns:
            Unique ID.
        """
        # this does NOT need to be cryptographically secure
        # it needs to be filesystem safe and *somewhat* unique
        if not self._uid:
            self._uid = sha1(
                str(self).encode(errors="replace"), usedforsecurity=False
            ).hexdigest()
        return self._uid


class Visit:
    """Visit contains details about the site and browser."""

    __slots__ = (
        "_end_time",
        "_start_time",
        "explorer",
        "idle_timestamp",
        "puppet",
        "url",
    )

    def __init__(self, puppet: FFPuppet, url: URL, explorer: Explorer | None) -> None:
        self._end_time: float | None = None
        self._start_time = perf_counter()
        self.explorer = explorer
        self.idle_timestamp: float | None = None
        self.puppet = puppet
        self.url = url

    def cleanup(self) -> None:
        """Close dependents and cleanup.

        Args:
            None

        Returns:
            None.
        """
        self.close()
        self.puppet.clean_up()

    def close(self) -> None:
        """Close browser and explorer if needed.

        Args:
            None

        Returns:
            None.
        """
        if self._end_time is None:
            self._end_time = perf_counter()
            # close browser before closing explorer
            self.puppet.close()
            if self.explorer is not None:
                LOG.debug("%s explorer: %s", self.url.uid[:6], self.explorer.state())
                self.explorer.close()

    def duration(self) -> float:
        """Total runtime of the visit.

        Args:
            None

        Returns:
            Visit duration.
        """
        if self._end_time is not None:
            return self._end_time - self._start_time
        return perf_counter() - self._start_time

    def is_active(self) -> bool:
        """Check if visit is in progress.

        Args:
            None

        Returns:
            True if visit is in progress otherwise False.
        """
        return self._end_time is None


# pylint: disable=too-many-instance-attributes
class SiteScout:
    """SiteScout can visit a collection of URLs and report process failures."""

    __slots__ = (
        "_active",
        "_binary",
        "_cert_files",
        "_complete",
        "_coverage",
        "_debugger",
        "_display",
        "_explore",
        "_extension",
        "_fuzzmanager",
        "_launch_failure_limit",
        "_launch_failures",
        "_launch_timeout",
        "_log_limit",
        "_memory_limit",
        "_prefs",
        "_profile",
        "_summaries",
        "_urls",
    )

    def __init__(
        self,
        binary: Path,
        profile: Path | None = None,
        prefs_js: Path | None = None,
        debugger: Debugger = Debugger.NONE,
        display: str | None = None,
        launch_timeout: int = 180,
        launch_failure_limit: int = 3,
        log_limit: int = 0,
        memory_limit: int = 0,
        extension: list[Path] | None = None,
        cert_files: list[Path] | None = None,
        explore: bool = False,
        fuzzmanager: bool = False,
        coverage: bool = False,
    ) -> None:
        assert launch_failure_limit > 0
        self._active: list[Visit] = []
        self._complete: list[Visit] = []
        self._summaries: list[VisitSummary] = []
        self._urls: list[URL] = []
        # browser related
        self._binary = binary
        self._cert_files = cert_files
        self._coverage = coverage
        self._debugger = debugger
        self._display = display
        self._explore = explore
        self._extension = extension
        self._launch_failure_limit = launch_failure_limit
        # consecutive launch failures
        self._launch_failures = 0
        self._launch_timeout = launch_timeout
        self._log_limit = log_limit
        self._memory_limit = memory_limit
        self._prefs = prefs_js
        self._profile = profile
        # reporter
        self._fuzzmanager: FuzzManagerReporter | None = (
            FuzzManagerReporter(binary, working_path=TMP_PATH) if fuzzmanager else None
        )

    def __enter__(self) -> SiteScout:
        return self

    def __exit__(self, *exc: object) -> None:
        self.close()

    def close(self) -> None:
        """Close and cleanup browser instances.

        Args:
            None

        Returns:
            None
        """
        LOG.debug("closing %d active visit(s)...", len(self._active))
        for visit in chain(self._active, self._complete):
            visit.cleanup()
        self._active.clear()
        self._complete.clear()

    def _launch(self, url: URL, log_path: Path | None = None) -> bool:
        """Launch a new browser instance and visit provided URL.

        Args:
            url: URL to visit.
            log_path: Directory to save launch failure logs.

        Returns:
            True if the browser was launched otherwise False.
        """
        env_mod: dict[str, str | None] = {"MOZ_CRASHREPORTER_SHUTDOWN": "1"}

        ffp = FFPuppet(
            debugger=self._debugger,
            headless=self._display,
            use_profile=self._profile,
            working_path=str(TMP_PATH),
        )
        try:
            ffp.launch(
                self._binary,
                env_mod=env_mod,
                location=None if self._explore else str(url),
                launch_timeout=self._launch_timeout,
                log_limit=self._log_limit,
                marionette=0 if self._explore else None,
                memory_limit=self._memory_limit,
                prefs_js=self._prefs,
                extension=self._extension,
                cert_files=self._cert_files,
            )
            self._launch_failures = 0
        except LaunchError as exc:
            self._launch_failures += 1
            is_failure = not isinstance(exc, BrowserTimeoutError)
            LOG.warning("Browser launch %s...", "failure" if is_failure else "timeout")
            if self._launch_failures >= self._launch_failure_limit:
                # save failure
                if is_failure and log_path is not None:
                    ffp.close()
                    dst = log_path / strftime("%Y%m%d-%H%M%S-launch-failure")
                    ffp.save_logs(dst)
                    LOG.warning("Logs saved '%s'", dst)
                raise

        finally:
            if self._launch_failures != 0:
                ffp.clean_up()

        if self._launch_failures == 0:
            explorer: Explorer | None = None
            if self._explore:
                assert ffp.marionette is not None
                explorer = Explorer(self._binary, ffp.marionette, str(url))
            self._active.append(Visit(ffp, url, explorer=explorer))

        return self._launch_failures == 0

    def load_dict(self, data: UrlDB) -> None:
        """Load URLs from a UrlDB and add them to the queue.

        Args:
            data: Dictionary containing URLs.

        Returns:
            None
        """
        total_domains = 0
        total_subdomains = 0
        total_urls = 0
        LOG.debug("processing dict...")
        existing = {x.uid for x in self._urls}
        while data:
            domain, subdomains = data.popitem()
            total_domains += 1
            for subdomain, paths in subdomains.items():
                total_subdomains += 1
                for path in paths:
                    url = URL(
                        domain,
                        subdomain=subdomain if subdomain != NO_SUBDOMAIN else None,
                        path=path,
                    )
                    # avoid duplicates
                    if url.uid not in existing:
                        self._urls.append(url)
                        existing.add(url.uid)
                    total_urls += 1
        LOG.debug(
            "%d domain(s), %d subdomain(s), %d URL(s) processed, %d loaded",
            total_domains,
            total_subdomains,
            total_urls,
            len(existing),
        )

    def load_str(self, url: str) -> None:
        """Parse, sanitize and add a URL to the queue.

        Args:
            url: Location to visit.

        Returns:
            None
        """
        assert url
        if "://" not in url:
            url = f"http://{url}"
        parsed = urlsplit(url, allow_fragments=False)
        if parsed.scheme not in URL.ALLOWED_SCHEMES:
            LOG.error("Unsupported scheme in URL: %r", url)
            return None
        path = parsed.path if parsed.path else "/"
        if parsed.query:
            path = f"{path}?{parsed.query}"
        # this currently does not separate domain and subdomain
        formatted = URL.create(parsed.netloc, path=path, scheme=parsed.scheme)
        # add unique urls to queue
        # NOTE: this might get slow with large lists
        if formatted is not None and formatted.uid not in {x.uid for x in self._urls}:
            self._urls.append(formatted)
        return None

    # pylint: disable=too-many-branches
    def _process_active(
        self,
        time_limit: float,
        idle_usage: int = 10,
        idle_wait: float = 10.0,
        min_visit: int = 20,
    ) -> None:
        """Check active browser processes and move complete visits from the active
        to the complete queue.

        Args:
            time_limit: Maximum time to wait for a site to load.
            idle_usage: Maximum CPU usage to be considered idle.
            idle_wait: Amount of time in seconds process must be idle.
            min_visit: Delay in seconds before idle checks begin.

        Returns:
            None
        """
        assert time_limit > 0
        assert min_visit > 0
        assert idle_usage >= 0
        assert idle_wait >= 0

        complete: list[int] = []
        for index, visit in enumerate(self._active):
            # check if work is complete
            visit_runtime = visit.duration()
            if not visit.puppet.is_healthy():
                visit.close()
                complete.append(index)
            elif visit_runtime >= time_limit:
                LOG.debug("visit timeout (%s)", visit.url.uid[:6])
                if self._coverage:
                    visit.puppet.dump_coverage()
                visit.close()
                complete.append(index)
            # check if explorer is complete but browser is running
            elif visit.explorer and not visit.explorer.is_running():
                LOG.debug("visit explorer not running (%s)", visit.url.uid[:6])
                if not visit.explorer.not_found():
                    # pause in case browser is closing
                    visit.puppet.wait(10)
                visit.close()
                complete.append(index)
            # check all browser processes are below idle limit
            elif idle_usage and visit_runtime >= min_visit:
                if all(x[1] < idle_usage for x in visit.puppet.cpu_usage()):
                    now = perf_counter()
                    if visit.idle_timestamp is None:
                        LOG.debug("set idle (%s)", visit.url.uid[:6])
                        visit.idle_timestamp = now
                    if now - visit.idle_timestamp >= idle_wait:
                        LOG.debug("visit idle (%s)", visit.url.uid[:6])
                        if self._coverage:
                            visit.puppet.dump_coverage()
                        visit.close()
                        complete.append(index)
                elif visit.idle_timestamp is not None:
                    LOG.debug("reset idle (%s)", visit.url.uid[:6])
                    visit.idle_timestamp = None

        if complete:
            LOG.debug("found %d complete", len(complete))
            for index in sorted(complete, reverse=True):
                self._complete.append(self._active.pop(index))
            LOG.debug("%d active, %d complete", len(self._active), len(self._complete))

    def _process_complete(self, log_path: Path) -> int:
        """Report results, record summaries and remove completed visits.

        Args:
            log_path: Directory to save results in.

        Returns:
            Number of results found.
        """
        results = 0
        while self._complete:
            LOG.debug("%d pending visit(s) to check", len(self._complete))
            visit = self._complete.pop()
            assert not visit.is_active()
            duration = visit.duration()
            summary = VisitSummary(duration, visit.url)

            if visit.puppet.reason in (Reason.ALERT, Reason.WORKER):
                summary.has_result = True
                dst = log_path / strftime(f"%Y%m%d-%H%M%S-result-{visit.url.uid[:6]}")
                visit.puppet.save_logs(dst)
                if self._prefs:
                    (dst / "prefs.js").write_text(self._prefs.read_text())
                (dst / "duration.txt").write_text(f"{duration:0.1f}")
                (dst / "url.txt").write_text(str(visit.url))
                results += 1
                LOG.info("Result found visiting '%s' (%0.1fs)", visit.url, duration)
                if self._fuzzmanager:
                    metadata = {
                        "duration": f"{duration:0.1f}",
                        "url": str(visit.url),
                    }
                    if visit.explorer is not None:
                        metadata["explore_state"] = visit.explorer.state()
                    fm_id, short_sig = self._fuzzmanager.submit(dst, metadata)
                    LOG.info("FuzzManager (%d): %s", fm_id, trim(short_sig, 60))
                    # remove local data when reporting to FM
                    rmtree(dst)
                else:
                    LOG.info("Saved as '%s'", dst)
            else:
                summary.force_closed = visit.puppet.reason != Reason.EXITED

            if visit.explorer is not None:
                summary.get_duration = visit.explorer.get_duration()
                summary.explore_duration = visit.explorer.explore_duration()
                summary.explore_state = visit.explorer.state()
                summary.not_found = visit.explorer.not_found()
                if summary.not_found:
                    LOG.info("Server Not Found: '%s'", visit.url)
                    self._skip_not_found(visit.url.domain)

            self._summaries.append(summary)
            visit.cleanup()
        return results

    def schedule_urls(
        self, url_limit: int = 0, randomize: bool = True, visits: int = 1
    ) -> None:
        """Prepare URL queue. Randomize and limit size as needed.

        Args:
            url_limit: Limit total URLs when value is greater than zero.
            shuffle_urls: Randomize order of URLs in the queue.
            visits: Number of times to visit each URL.

        Returns:
            None
        """
        assert url_limit >= 0
        if randomize:
            # shuffle before enforcing limit so all entries can potentially be included
            shuffle(self._urls)
        if url_limit and len(self._urls) > url_limit:
            LOG.info("Enforcing URL limit (%d -> %d)", len(self._urls), url_limit)
            self._urls = self._urls[:url_limit]
        if not randomize:
            # provided url list is processed in reverse order
            # reverse the list to maintain original order
            self._urls.reverse()
        if visits > 1:
            LOG.info("Each URL will be visited %dx", visits)
            # repeat the list for multiple visits
            self._urls = self._urls * visits

    def _skip_not_found(self, domain: str) -> None:
        """Remove URLs with matching domain from the queue.

        Args:
            domain: Value used as filter.

        Returns:
            None.
        """
        removed = 0
        for idx in reversed(range(len(self._urls))):
            if self._urls[idx].domain == domain:
                url = self._urls.pop(idx)
                self._summaries.append(VisitSummary(0, url, not_found=True))
                removed += 1
        if removed > 0:
            LOG.info("Skipping %d related queued URLs", removed)

    def _skip_remaining(self) -> None:
        """Skip remaining visits. Clear the URL queue, close and cleanup active browser
        instances.

        Args:
            None

        Returns:
            None
        """
        self._urls.clear()
        LOG.debug("skipping active visits: %d", len(self._active))
        for visit in self._active:
            visit.cleanup()
        self._active.clear()

    # pylint: disable=too-many-locals,too-many-statements
    def run(
        self,
        log_path: Path,
        time_limit: int,
        check_delay: int = 1,
        domain_rate_limit: int = 20,
        instance_limit: int = 1,
        status_report: Path | None = None,
        result_limit: int = 0,
        runtime_limit: int = 0,
    ) -> None:
        """Iterate over the queue and visit each URL. Each visit is performed by a
        new browser instance with a clean profile.

        Args:
            log_path: Location to write results.
            time_limit: Maximum time in seconds of a site visit.
            check_delay: Time in seconds between checking for results.
            domain_rate_limit: Minimum time in seconds between visiting the same domain.
            instance_limit: Maximum number of browser sessions to run at once.
            status_report: File to populate with status report data.
            result_limit: Number of results that can be found before exiting.
            runtime_limit: Maximum number of seconds to run.

        Returns:
            None
        """
        assert check_delay >= 0
        assert domain_rate_limit >= 0
        assert instance_limit > 0
        assert result_limit >= 0
        assert runtime_limit >= 0

        if runtime_limit > 0:
            minutes, seconds = divmod(runtime_limit, 60)
            hours, minutes = divmod(minutes, 60)
            LOG.info("Runtime limit is %02d:%02d:%02d", hours, minutes, seconds)

        end_time = int(perf_counter() + runtime_limit) if runtime_limit > 0 else 0
        last_visit: dict[str, float] = {}
        status = Status(status_report) if status_report else None
        total_results = 0
        total_urls = len(self._urls)
        while self._urls or self._active:
            # perform status report
            if status:
                status.report(
                    len(self._active),
                    instance_limit,
                    len(self._summaries),
                    total_urls,
                    total_results,
                    sum(1 for x in self._summaries if x.not_found),
                    0,
                )

            # select url to visit and launch browser
            if self._urls and (len(self._active) < instance_limit):
                next_url = self._urls.pop()
                # avoid frequent domain visits by rate limiting
                if (
                    next_url.domain in last_visit
                    and perf_counter() - last_visit[next_url.domain] < domain_rate_limit
                ):
                    LOG.debug("domain rate limit hit (%s)", next_url.domain)
                    # move url to the end of the queue
                    self._urls.insert(0, next_url)
                # attempt to launch browser and visit url
                elif self._launch(next_url, log_path=log_path):
                    short_url = trim(str(next_url), 80)
                    LOG.info(
                        "[%02d/%02d] %r",
                        len(self._active) + len(self._summaries),
                        total_urls,
                        short_url,
                    )
                    LOG.debug(
                        "launched, explore: %r, timeout: %ds, %s - %s",
                        self._explore,
                        time_limit,
                        next_url.uid[:6],
                        short_url,
                    )
                    last_visit[next_url.domain] = perf_counter()
                    assert self._active
                # launch failed
                else:
                    # re-add url to queue since it was not visited
                    self._urls.append(next_url)

            # check for complete processes (disable idle checks when explore is set)
            self._process_active(time_limit, idle_usage=0 if self._explore else 10)
            total_results += self._process_complete(log_path=log_path)

            # check result and runtime limits
            if result_limit and self._urls and total_results >= result_limit:
                LOG.info("Result limit (%d) hit", result_limit)
                self._skip_remaining()
                assert not self._active
                assert not self._urls
            elif 0 < end_time <= perf_counter():
                LOG.info("Runtime limit (%ds) hit", runtime_limit)
                self._skip_remaining()
                assert not self._active
                assert not self._urls

            # wait a moment for work to complete
            if self._urls or self._active:
                sleep(check_delay)

        # final status report
        if status:
            if self._summaries:
                # don't include "server not found" results in calculation
                avg_duration = int(
                    sum(x.duration for x in self._summaries if not x.not_found)
                    / len(self._summaries)
                )
            else:
                avg_duration = 0

            status.report(
                len(self._active),
                instance_limit,
                len(self._summaries),
                total_urls,
                total_results,
                sum(1 for x in self._summaries if x.not_found),
                avg_duration,
                force=True,
            )
        LOG.info(
            "Visits complete: %d, results: %d", len(self._summaries), total_results
        )


# pylint: disable=too-many-return-statements
def verify_dict(data: Any, allow_empty: bool = False) -> str | None:
    """Verify the structure of data.

    Args:
        data: Dictionary to check.
        allow_empty: Empty data set is valid if True.

    Return:
        An error message is returned if a problem is found.
    """
    if not isinstance(data, dict):
        return "Invalid data"
    if not data and not allow_empty:
        return "No data found"
    # check domains
    for domain, subdomains in data.items():
        if not isinstance(domain, str) or not domain:
            return "Domain must be a string"
        if not isinstance(subdomains, dict) or not subdomains:
            return f"Invalid domain entry: '{domain}'"
        # check subdomains
        for subdomain, paths in subdomains.items():
            if not isinstance(subdomain, str) or not subdomain:
                return "Subdomain must be a string"
            if not isinstance(paths, list) or not paths:
                return f"Invalid subdomain entry: '{subdomain}' in '{domain}'"
            # check paths
            for path in paths:
                if not isinstance(path, str) or not path.startswith("/"):
                    return "Path must be a string starting with '/'"
    return None



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=missing-docstring,protected-access
from pytest import mark, raises

from .url_collection import UrlCollection, main, parse_args

BASIC_URL_DB = {
    "d1": {"*": ["/", "/d1-path1"]},
    "d2": {
        "sd2-1": ["/sd2-path1"],
        "sd2-2": ["/"],
    },
}


@mark.parametrize(
    "url_str, expected",
    [
        ("http://domain.com/", "http://domain.com/"),
        ("http://sub.domain.com/", "http://sub.domain.com/"),
        ("https://sub.domain.com/", "http://sub.domain.com/"),
        ("sub.domain.com", "http://sub.domain.com/"),
        ("domain.com", "http://domain.com/"),
        ("domain.com/ ", "http://domain.com/"),
        ("domain.com/test ", "http://domain.com/test"),
        ("user:pass@domain.com", "http://domain.com/"),
        ("domain.co.uk", "http://domain.co.uk/"),
        ("domain.com/test", "http://domain.com/test"),
        ("domain.com/test?123", "http://domain.com/test?123"),
        ("domain.com/test#123", "http://domain.com/test#123"),
        ("HTTP://SUB.DOMAIN.COM/TEST", "http://sub.domain.com/TEST"),
        ("domain.com/El Niño", "http://domain.com/El%20Ni%C3%B1o"),
        ("fóò.café.fr/test", "http://xn--f-tgac.xn--caf-dma.fr/test"),
        ("*.domain.com", None),
        ("failed-parse", None),
    ],
)
def test_url_collection_parse_url(url_str, expected):
    """test UrlCollection._parse_url()"""
    parsed = UrlCollection._parse_url(url_str, default_subdomain=None)
    if expected is None:
        assert parsed is None
    else:
        assert parsed
        assert str(parsed) == expected


def test_url_collection_01():
    """test empty UrlCollection()"""
    urls = UrlCollection()
    assert urls is not None
    assert not urls
    assert len(urls) == 0
    assert not tuple(urls.domains)


def test_url_collection_02(tmp_path):
    """test simple UrlCollection()"""
    urls = UrlCollection(BASIC_URL_DB)
    assert urls
    assert len(urls) == 4
    assert len(tuple(urls.domains)) == 2
    # save and load
    yml = tmp_path / "urls.yml"
    urls.save_yml(yml)
    urls = UrlCollection.load_yml(yml)
    assert urls
    # add/remove urls
    assert len(urls) == 4
    assert urls.add_url("test.com")
    assert len(urls) == 5
    assert urls.remove_url("test.com")
    assert not urls.remove_url("test.com")
    # try adding a urls that can't be parsed
    assert not urls.unparsable
    assert not urls.add_url("invalid")
    assert "invalid" in urls.unparsable
    # add a list of urls from a file
    txt = tmp_path / "urls_list.txt"
    txt.write_text("1.test.com\n\n2.test.com\n#ignored.com\nother.com")
    assert urls.add_list(txt)
    assert not urls.add_list(txt)
    assert len(urls) == 7


def test_url_collection_sorting():
    """test UrlCollection() sorting"""
    urls = UrlCollection({"a.com": {"b": ["/z", "/", "/b"]}})
    assert urls._db["a.com"]["b"][0] == "/z"
    assert urls._db["a.com"]["b"][1] == "/"
    urls.sort_paths()
    assert urls._db["a.com"]["b"][0] == "/"
    assert urls._db["a.com"]["b"][1] == "/b"
    # added paths should be sorted
    urls.add_url("b.a.com/a")
    assert urls._db["a.com"]["b"][0] == "/"
    assert urls._db["a.com"]["b"][1] == "/a"
    assert urls._db["a.com"]["b"][2] == "/b"


def test_parse_args(capsys, tmp_path):
    """test parse_args()"""
    #
    data_file = tmp_path / "test.yml"
    args = parse_args([str(data_file)])
    assert args.url_db == data_file

    #
    with raises(SystemExit):
        parse_args([str(data_file), "-l", "missing_list.txt"])
    assert "error: missing_list.txt is not a file." in capsys.readouterr().err


def test_main(tmp_path):
    """test main()"""

    # no data file
    data_file = tmp_path / "test.yml"
    assert main([str(data_file)]) == 0
    assert not data_file.exists()

    # invalid data file
    data_file.write_bytes(b"\0")
    assert main([str(data_file)]) == 1
    data_file.unlink()

    # remove url (data file does not exists)
    assert main([str(data_file), "-r", "test.com"]) == 0
    assert not data_file.exists()

    # add single url
    assert main([str(data_file), "-u", "test.com", "-d"]) == 0
    assert data_file.exists()

    # remove url (data file now exists)
    file_size = data_file.stat().st_size
    assert main([str(data_file), "-r", "test.com"]) == 0
    assert data_file.stat().st_size < file_size
    assert data_file.exists()

    # add unparsable
    file_size = data_file.stat().st_size
    assert main([str(data_file), "-u", "blah"]) == 0
    assert data_file.stat().st_size == file_size

    # remove unparsable
    file_size = data_file.stat().st_size
    assert main([str(data_file), "-r", "blah"]) == 0
    assert data_file.stat().st_size == file_size

    # add urls from list file
    txt = tmp_path / "urls_list.txt"
    txt.write_text("1.test.com\n\n2.test.com\n#ignored.com\nother.com")
    file_size = data_file.stat().st_size
    assert main([str(data_file), "-l", str(txt)]) == 0
    assert data_file.stat().st_size > file_size



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from enum import Enum, auto, unique
from logging import ERROR, WARNING, getLogger
from threading import Event, Lock, Thread
from time import perf_counter
from typing import TYPE_CHECKING

from page_explorer import ExplorerError, PageExplorer

if TYPE_CHECKING:
    from pathlib import Path

# suppress noise from page explorer deps
getLogger("selenium").setLevel(WARNING)
getLogger("urllib3").setLevel(ERROR)

LOG = getLogger(__name__)


@unique
class State(Enum):
    """State of Explorer"""

    PENDING = auto()
    CONNECTING = auto()
    NOT_FOUND = auto()
    LOADING = auto()
    EXPLORING = auto()
    CLOSING = auto()
    CLOSED = auto()


class ExplorerStatus:
    """Used to collect status data from PageExplorer."""

    __slots__ = ("explore_duration", "get_duration", "lock", "state")

    def __init__(self) -> None:
        self.explore_duration: float | None = None
        self.get_duration: float | None = None
        self.lock = Lock()
        self.state: State = State.PENDING


class Explorer:
    """PageExplorer wrapper to use enable use via SiteScout."""

    __slots__ = ("_can_skip", "_status", "_thread")

    def __init__(self, binary: Path, port: int, url: str) -> None:
        assert port is not None
        self._can_skip = Event()
        self._status = ExplorerStatus()
        self._thread = Thread(
            target=self._run,
            args=(binary, port, url, self._status, self._can_skip),
        )
        self._thread.start()

    def __enter__(self) -> Explorer:
        return self

    def __exit__(self, *exc: object) -> None:
        self.close()

    def close(self) -> None:
        """Unblock and join PageExplorer thread.

        Args:
            None

        Returns:
            None
        """
        self._can_skip.set()
        # this should NEVER timeout, if it does there is a bug
        self._thread.join(timeout=60)

    def explore_duration(self) -> float | None:
        """Amount of time in seconds spent exploring content.

        Args:
            None

        Returns:
            Time in seconds if available otherwise None.
        """
        with self._status.lock:
            return self._status.explore_duration

    def get_duration(self) -> float | None:
        """Amount of time in seconds spent loading content.

        Args:
            None

        Returns:
            Time in seconds if available otherwise None.
        """
        with self._status.lock:
            return self._status.get_duration

    def is_running(self) -> bool:
        """Check if PageExplorer thread is running.

        Args:
            None

        Returns:
            True if the thread is running otherwise False.
        """
        return self._thread.is_alive()

    def not_found(self) -> bool:
        """Check if server did not respond (server not found).

        Args:
            None

        Returns:
            True if the server did not respond otherwise False.
        """
        with self._status.lock:
            return self._status.state == State.NOT_FOUND

    def state(self) -> str:
        """Current state of the explorer.

        Args:
            None

        Returns:
            Current state.
        """
        with self._status.lock:
            return self._status.state.name

    @staticmethod
    def _run(
        binary: Path,
        port: int,
        url: str,
        status: ExplorerStatus,
        can_skip: Event,
    ) -> None:
        """Interact with content via PageExplorer.

        Args:
            binary: Binary launched.
            port: Browser driver listening port.
            url: Url to visit.
            status: Status tracker.
            can_skip: Used to delay execution.

        Returns:
            None.
        """

        def custom_wait(seconds: float) -> None:  # pragma: no cover
            assert seconds >= 0
            can_skip.wait(seconds)

        with status.lock:
            status.state = State.CONNECTING

        try:
            with PageExplorer(binary=binary, port=port) as explorer:
                with status.lock:
                    status.state = State.LOADING
                start_time = perf_counter()
                if not explorer.get(url):
                    # failed to navigate to web site
                    if explorer.is_connected():
                        with status.lock:
                            status.state = State.NOT_FOUND
                    return
                duration = perf_counter() - start_time
                with status.lock:
                    status.get_duration = duration
                    status.state = State.EXPLORING
                start_time = perf_counter()
                if not explorer.explore(wait_cb=custom_wait):
                    # failed to execute all explore instructions
                    return
                duration = perf_counter() - start_time
                with status.lock:
                    status.explore_duration = duration
                    status.state = State.CLOSING
                explorer.close_browser()
                with status.lock:
                    status.state = State.CLOSED
        except ExplorerError:
            LOG.debug("ExplorerError detected, aborting...")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from bisect import insort
from logging import DEBUG, INFO, getLogger
from pathlib import Path
from typing import TYPE_CHECKING, cast
from urllib.parse import urlsplit

from tldextract import extract
from yaml import dump

try:
    from yaml import CSafeDumper as SafeDumper
except ImportError:
    from yaml import SafeDumper  # type: ignore

from .main import init_logging, load_input
from .site_scout import NO_SUBDOMAIN, URL, UrlDB

if TYPE_CHECKING:
    from collections.abc import Generator

LOG = getLogger(__name__)


class UrlCollection:
    """Manage collections of URLs."""

    __slots__ = ("_db", "unparsable")

    def __init__(self, url_db: UrlDB | None = None) -> None:
        self._db = url_db or cast(UrlDB, {})
        self.unparsable: set[str] = set()

    def __len__(self) -> int:
        count = 0
        for domain in self._db:
            for sub in self._db[domain]:
                count += len(self._db[domain][sub])
        return count

    def __iter__(self) -> Generator[URL]:
        for domain in self._db:
            for sub in self._db[domain]:
                for path in self._db[domain][sub]:
                    yield URL(domain, subdomain=sub, path=path)

    @staticmethod
    def _parse_url(url: str, default_subdomain: str = NO_SUBDOMAIN) -> URL | None:
        """Parse URL from given string.

        Args:
            url: URL to parse.
            default_subdomain: Placeholder for empty subdomain.

        Returns:
            URL object if successfully parsed otherwise None.
        """
        if "://" not in url:
            url = f"http://{url}"
        parsed_url = urlsplit(url, allow_fragments=False)
        # TODO: handle params?
        assert parsed_url.netloc
        # URL domain info
        udi = extract(parsed_url.netloc)
        if not udi.domain or not udi.suffix or udi.subdomain == NO_SUBDOMAIN:
            return None
        assert udi.domain, f"{parsed_url.netloc} -> {udi}"
        assert udi.suffix, f"{parsed_url.netloc} -> {udi}"
        path = parsed_url.path.strip() or "/"
        if parsed_url.query:
            path = f"{path}?{parsed_url.query}"
        return URL.create(
            f"{udi.domain}.{udi.suffix}",
            # replace empty/missing subdomain with placeholder
            subdomain=udi.subdomain if udi.subdomain else default_subdomain,
            path=path,
        )

    def add_list(self, urls_file: Path) -> int:
        """Load a text file containing a line separated list of URLs and add them
        to the collection of known URLs.

        Args:
            urls_file: File to load.

        Returns:
            Number of previously unknown URLs that have been added.
        """
        added = 0
        with urls_file.open() as in_fp:
            for line in in_fp:
                entry = line.strip()
                if not entry or entry[0] == "#":
                    # skip comments and empty lines
                    continue
                if self.add_url(entry) is not None:
                    added += 1
        return added

    def add_url(self, url: str) -> URL | None:
        """Load a single URL and add it to the list of known URLs.

        Args:
            url: URL to add.

        Returns:
            URL that was added.
        """
        parsed = self._parse_url(url)
        if parsed is None:
            LOG.debug("failed to parse and add: %r", url)
            self.unparsable.add(url)
        else:
            assert parsed.subdomain is not None
            if parsed.domain not in self._db:
                self._db[parsed.domain] = {}
            if parsed.subdomain not in self._db[parsed.domain]:
                self._db[parsed.domain][parsed.subdomain] = []
            if parsed.path not in self._db[parsed.domain][parsed.subdomain]:
                insort(self._db[parsed.domain][parsed.subdomain], parsed.path)
                LOG.debug("added: %s", parsed)
                return parsed
        return None

    @property
    def domains(self) -> Generator[str]:
        """All known domains.

        Args:
            None

        Yields:
            Domains in URL collection.
        """
        yield from self._db

    def remove_url(self, url: str) -> bool:
        """Remove a URL from the list of known URLs.

        Args:
            url: URL to remove.

        Returns:
            True if the URL was removed otherwise False.
        """
        parsed = self._parse_url(url)
        if parsed is None:
            LOG.debug("failed to parse and remove: %r", url)
            return False
        assert parsed.subdomain is not None

        try:
            self._db[parsed.domain][parsed.subdomain].remove(parsed.path)
        except (KeyError, ValueError):
            LOG.debug("failed to remove unknown url: %r", url)
            return False

        # remove empty domain/subdomain entries
        if not self._db[parsed.domain][parsed.subdomain]:
            del self._db[parsed.domain][parsed.subdomain]
        if not self._db[parsed.domain]:
            del self._db[parsed.domain]

        return True

    @classmethod
    def load_yml(cls, src: Path) -> UrlCollection | None:
        """Load UrlCollection from a YML file.

        Args:
            src: File to load.

        Returns:
            UrlCollection.
        """
        data = next(load_input([src], allow_empty=True), None)
        if data is not None:
            return cls(data)
        return None

    def save_yml(self, dst: Path) -> None:
        """Save UrlCollection to a YML file.

        Args:
            dst: File to save to.

        Returns:
            None
        """
        with dst.open("w") as out_fp:
            dump(self._db, out_fp, Dumper=SafeDumper, default_style="'")

    def sort_paths(self) -> None:
        """Sort paths.

        Args:
            None

        Returns:
            None.
        """
        for subdomains in self._db.values():
            for paths in subdomains.values():
                paths.sort()


def parse_args(argv: list[str] | None = None) -> Namespace:
    """Argument parsing"""
    parser = ArgumentParser(description="Manage URL collections used by Site-scout.")
    parser.add_argument("url_db", type=Path, help="YML file containing data.")
    parser.add_argument("-d", "--display-list", action="store_true")
    level_map = {"INFO": INFO, "DEBUG": DEBUG}
    parser.add_argument(
        "--log-level",
        choices=sorted(level_map),
        default="INFO",
        help="Configure console logging (default: %(default)s).",
    )
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument(
        "-l",
        "--url-list",
        type=Path,
        help="File containing line separated list of URLs.",
    )
    group.add_argument("-r", "--remove", help="Remove single URL.")
    group.add_argument("-u", "--url", help="Add single URL.")
    args = parser.parse_args(argv)

    args.log_level = level_map[args.log_level]

    if args.url_list and not args.url_list.is_file():
        parser.error(f"{args.url_list} is not a file.")

    return args


# pylint: disable=too-many-branches
def main(argv: list[str] | None = None) -> int:
    """Main function"""
    args = parse_args(argv)
    init_logging(args.log_level)

    try:
        # load data
        if args.url_db and args.url_db.is_file():
            LOG.info("Loading '%s'...", args.url_db)
            urls = UrlCollection.load_yml(args.url_db)
            if urls is None:
                return 1
            urls.sort_paths()
        else:
            urls = UrlCollection()

        # update data
        changed = False
        if args.url:
            added = urls.add_url(args.url)
            if added is not None:
                LOG.info("Added '%s'", added)
                changed = True
        elif args.url_list:
            LOG.info("Scanning URL(s) in '%s'...", args.url_list)
            added_count = urls.add_list(args.url_list)
            LOG.info("Added %s URL(s).", f"{added_count:,d}")
            changed = added_count > 0
        elif args.remove:
            changed = urls.remove_url(args.remove)
            if changed:
                LOG.info("Removed 1 URL.")

        # save data
        if changed:
            LOG.info("Saving '%s'...", args.url_db)
            urls.save_yml(args.url_db)
        else:
            LOG.info("No change made.")

        # show current list of URLs
        if args.display_list:
            LOG.info("Known URLs: %s", f"{len(urls):,d}")
            for count, entry in enumerate(urls, start=1):
                LOG.info("%04d: %s", count, entry)

        # show unparsable input
        if urls.unparsable:
            LOG.warning("Unparsable entries: %s", f"{len(urls.unparsable):,d}")
            for unparsable in urls.unparsable:
                LOG.info("- '%s'", unparsable)

        # show summary
        try:
            file_size = args.url_db.stat().st_size
        except FileNotFoundError:
            file_size = 0
        LOG.info(
            "%s URLs (%s domains) in db (%s bytes).",
            f"{len(urls):,d}",
            f"{sum(1 for _ in urls.domains):,d}",
            f"{file_size:,d}",
        )

    except KeyboardInterrupt:  # pragma: no cover
        LOG.warning("Aborting...")

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



# type: ignore
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from pytest import mark

from .explorer import Explorer, ExplorerError, State


@mark.parametrize(
    "state, get_return, explore_return",
    (
        (State.NOT_FOUND, False, False),
        (State.EXPLORING, True, False),
        (State.CLOSED, True, True),
    ),
)
def test_explorer(mocker, tmp_path, state, get_return, explore_return):
    """test Explorer()"""
    page_explorer = mocker.patch(
        "site_scout.explorer.PageExplorer", autospec=True
    ).return_value.__enter__.return_value
    page_explorer.get.return_value = get_return
    page_explorer.explore.return_value = explore_return
    with Explorer(tmp_path, 0, "http://foo.foo") as explorer:
        assert not explorer._can_skip.is_set()
        explorer.close()
        assert not explorer.is_running()
        assert explorer._can_skip.is_set()
        assert explorer.state() == state.name
        assert explorer.not_found() == (state == State.NOT_FOUND)
        if get_return:
            assert explorer.get_duration() > 0
        else:
            assert explorer.get_duration() is None
        if explore_return:
            assert explorer.explore_duration() > 0
        else:
            assert explorer.explore_duration() is None


def test_explorer_failed(mocker, tmp_path):
    """test Explorer() failed to create PageExplorer"""
    mocker.patch("site_scout.explorer.PageExplorer", side_effect=ExplorerError("test"))
    with Explorer(tmp_path, 0, "http://foo.foo") as explorer:
        assert explorer.state() == State.CONNECTING.name
        assert explorer.get_duration() is None
        assert explorer.explore_duration() is None
